import util
from util import StorageServiceUtil, ENVIRONMENT_MAP, generate_volume_label
import mocker
from boto.exception import NoAuthHandlerFound, EC2ResponseError
from socket import gaierror
import subprocess
from testing import TestHookenv


class TestNovaUtil(mocker.MockerTestCase):

    def setUp(self):
        super(TestNovaUtil, self).setUp()
        self.maxDiff = None
        util.hookenv = TestHookenv(
            {"key": "myusername", "tenant": "myusername_project",
             "secret": "password", "region": "region1",
             "endpoint": "https://keystone_url:443/v2.0/",
             "default_volume_size": 11})
        util.log = util.hookenv.log
        self.storage = StorageServiceUtil("nova")

    def _set_environment_vars(self, environ={}):
        self.addCleanup(setattr, util.os, "environ", util.os.environ)
        util.os.environ = environ

    def test_invalid_provier_config(self):
        """When an invalid provider config is set and error is reported."""
        result = self.assertRaises(SystemExit, StorageServiceUtil, "ce2")
        self.assertEqual(result.code, 1)
        message = (
            "ERROR: Invalid charm configuration setting for provider. "
            "'ce2' must be one of: %s" % ", ".join(ENVIRONMENT_MAP.keys()))
        self.assertIn(
            message, util.hookenv._log_ERROR, "Not logged- %s" % message)

    def test_generate_volume_label(self):
        """
        L{generate_volume_label} return a string based on the C{remove_unit}
        """
        self.assertEqual("blah unit volume", generate_volume_label("blah"))

    def test_load_environment_with_nova_variables(self):
        """
        L{load_environment} will setup script environment variables for nova
        by mapping configuration values provided to openstack OS_* environment
        variables and then call L{validate_credentials} to assert
        that environment variables provided give access to the service.
        """
        self._set_environment_vars({})

        def mock_validate():
            pass
        self.storage.validate_credentials = mock_validate

        self.storage.load_environment()
        expected = {
            "OS_AUTH_URL": "https://keystone_url:443/v2.0/",
            "OS_PASSWORD": "password",
            "OS_REGION_NAME": "region1",
            "OS_TENANT_NAME": "myusername_project",
            "OS_USERNAME": "myusername"
        }
        self.assertEqual(util.os.environ, expected)

    def test_load_environment_error_missing_config_options(self):
        """
        L{load_environment} will exit in failure and log a message if any
        required configuration option is not set.
        """
        self._set_environment_vars({})

        def mock_validate():
            raise SystemExit("something invalid")
        self.storage.validate_credentials = mock_validate

        self.assertRaises(SystemExit, self.storage.load_environment)

    def test_validate_credentials_failure(self):
        """
        L{validate_credentials} will attempt a simple nova command to ensure
        the environment is properly configured to access the nova service.
        Upon failure to contact the nova service, L{validate_credentials} will
        exit in error and log a message.
        """
        command = "nova list"
        nova_cmd = self.mocker.replace(subprocess.check_call)
        nova_cmd(command, shell=True)
        self.mocker.throw(subprocess.CalledProcessError(1, command))
        self.mocker.replay()

        result = self.assertRaises(
            SystemExit, self.storage.validate_credentials)
        self.assertEqual(result.code, 1)
        message = (
            "ERROR: Charm configured credentials can't access nova endpoint. "
            "Command '%s' returned non-zero exit status 1" % command)
        self.assertIn(
            message, util.hookenv._log_ERROR, "Not logged- %s" % message)

    def test_validate_credentials(self):
        """
        L{validate_credentials} will succeed when a simple nova command
        succeeds due to a properly configured environment based on the charm
        configuration options.
        """
        command = "nova list"
        nova_cmd = self.mocker.replace(subprocess.check_call)
        nova_cmd(command, shell=True)
        self.mocker.replay()

        self.storage.validate_credentials()
        message = (
            "Validated charm configuration credentials have access to "
            "block storage service"
        )
        self.assertIn(
            message, util.hookenv._log_INFO, "Not logged- %s" % message)

    def test_wb_nova_volume_show_with_attachments(self):
        """
        L{_nova_volume_show} returns a C{dict} of volume information for
        the given C{volume_id}.
        """
        volume_id = "123-123-123"
        command = "nova volume-show '%s'" % volume_id
        output = (
            "+-------------------------------------+\n"
            "| Property | Value |\n"
            "+---------------------------------------------------------+\n"
            "| status              | in-use                            |\n"
            "| display_name        | my-volume                         |\n"
            "| attachments         | [{u'device': u'/dev/vdc', "
            "u'server_id': u'blah', u'id': u'%s', u'volume_id': u'%s'}] |\n"
            "| availability_zone   | nova                              |\n"
            "| bootable            | false                             |\n"
            "| created_at          | 2014-02-12T21:02:29.000000        |\n"
            "| display_description | None                              |\n"
            "| volume_type         | None                              |\n"
            "| snapshot_id         | None                              |\n"
            "| source_volid        | None                              |\n"
            "| size                | 9                                 |\n"
            "| id                  | %s                                |\n"
            "| metadata            | {}                                |\n"
            "+---------------------+-----------------------------------+\n" %
            (volume_id, volume_id, volume_id))
        nova_cmd = self.mocker.replace(subprocess.check_output)
        nova_cmd(command, shell=True)
        self.mocker.result(output)
        self.mocker.replay()

        expected = {
            "device": "/dev/vdc", "instance_id": "blah", "id": volume_id,
            "size": "9", "volume_label": "my-volume",
            "snapshot_id": "None", "availability_zone": "nova",
            "status": "in-use", "tags": {"volume_label": "my-volume"}}

        self.assertEqual(
            self.storage._nova_volume_show(volume_id), expected)

    def test_wb_nova_volume_show_without_attachments(self):
        """
        L{_nova_volume_show} returns a C{dict} of volume information for
        the given C{volume_id}. When no attachments are present, C{device} and
        C{instance_id} will be an empty C{str}.
        """
        volume_id = "123-123-123"
        command = "nova volume-show '%s'" % volume_id
        output = (
            "+-------------------------------------+\n"
            "| Property | Value |\n"
            "+---------------------------------------------------------+\n"
            "| status              | in-use                            |\n"
            "| display_name        | my-volume                         |\n"
            "| attachments         | []                                |\n"
            "| availability_zone   | nova                              |\n"
            "| bootable            | false                             |\n"
            "| created_at          | 2014-02-12T21:02:29.000000        |\n"
            "| display_description | None                              |\n"
            "| volume_type         | None                              |\n"
            "| snapshot_id         | None                              |\n"
            "| source_volid        | None                              |\n"
            "| size                | 9                                 |\n"
            "| id                  | %s                                |\n"
            "| metadata            | {}                                |\n"
            "+---------------------+-----------------------------------+\n" %
            (volume_id))
        nova_cmd = self.mocker.replace(subprocess.check_output)
        nova_cmd(command, shell=True)
        self.mocker.result(output)
        self.mocker.replay()

        expected = {
            "device": "", "instance_id": "", "id": volume_id,
            "size": "9", "volume_label": "my-volume",
            "snapshot_id": "None", "availability_zone": "nova",
            "status": "in-use", "tags": {"volume_label": "my-volume"}}

        self.assertEqual(
            self.storage._nova_volume_show(volume_id), expected)

    def test_wb_nova_volume_show_no_volume_present(self):
        """
        L{_nova_volume_show} exits when no matching volume is present due to
        receiving a C{CalledProcessError} and logs the error when unable to
        execute the C{nova volume-show} command.
        """
        volume_id = "123-123-123"
        command = "nova volume-show '%s'" % volume_id
        nova_cmd = self.mocker.replace(subprocess.check_output)
        nova_cmd(command, shell=True)
        self.mocker.throw(subprocess.CalledProcessError(1, command))
        self.mocker.replay()

        result = self.assertRaises(
            SystemExit, self.storage._nova_volume_show, volume_id)
        self.assertEqual(result.code, 1)
        message = (
            "ERROR: Failed to get nova volume info. Command '%s' returned "
            "non-zero exit status 1" % (command))
        self.assertIn(
            message, util.hookenv._log_ERROR, "Not logged- %s" % message)

    def test_get_volume_id_by_volume_name(self):
        """
        L{get_volume_id} provided with a existing C{volume_name} returns the
        corresponding nova volume id.
        """
        volume_name = "my-volume"
        volume_id = "12312412-412312"

        def mock_describe():
            return {volume_id: {"tags": {"volume_name": volume_name}},
                    "456456-456456": {"tags": {"volume_name": "blah"}}}
        self.storage.describe_volumes = mock_describe

        self.assertEqual(self.storage.get_volume_id(volume_name), volume_id)

    def test_get_volume_id_without_volume_name(self):
        """
        L{get_volume_id} without a provided C{volume_name} will discover the
        nova volume id by searching L{describe_volumes} for volumes labelled
        with the os.environ[JUJU_REMOTE_UNIT].
        """
        unit_name = "postgresql/0"
        self._set_environment_vars({"JUJU_REMOTE_UNIT": unit_name})
        volume_id = "123134124-1241412-1242141"

        def mock_describe():
            return {volume_id:
                    {"tags": {"volume_name": "postgresql/0 unit volume"}},
                    "456456-456456": {"tags": {"volume_name": "blah"}}}
        self.storage.describe_volumes = mock_describe

        self.assertEqual(self.storage.get_volume_id(), volume_id)

    def test_get_volume_id_without_volume_name_no_matching_volume(self):
        """
        L{get_volume_id} without a provided C{volume_name} will return C{None}
        when it cannot find a matching volume label from L{describe_volumes}
        for the os.environ[JUJU_REMOTE_UNIT].
        """
        unit_name = "postgresql/0"
        self._set_environment_vars({"JUJU_REMOTE_UNIT": unit_name})

        def mock_describe(val):
            self.assertIsNone(val)
            return {"123123-123123":
                    {"tags": {"volume_name": "postgresql/1 unit volume"}},
                    "456456-456456": {"tags": {"volume_name": "blah"}}}
        self.storage._nova_describe_volumes = mock_describe

        self.assertIsNone(self.storage.get_volume_id())

    def test_get_volume_id_without_volume_name_multiple_matching_volumes(self):
        """
        L{get_volume_id} does not support multiple volumes associated with the
        the instance represented by os.environ[JUJU_REMOTE_UNIT]. When
        C{volume_name} is not specified and L{describe_volumes} returns
        multiple results the function exits with an error.
        """
        unit_name = "postgresql/0"
        self._set_environment_vars({"JUJU_REMOTE_UNIT": unit_name})

        def mock_describe():
            return {"123123-123123":
                    {"tags": {"volume_name": "postgresql/0 unit volume"}},
                    "456456-456456":
                    {"tags": {"volume_name": "unit postgresql/0 volume2"}}}
        self.storage.describe_volumes = mock_describe

        result = self.assertRaises(SystemExit, self.storage.get_volume_id)
        self.assertEqual(result.code, 1)
        message = (
            "Error: Multiple volumes are associated with %s. "
            "Cannot get_volume_id." % unit_name)
        self.assertIn(
            message, util.hookenv._log_ERROR, "Not logged- %s" % message)

    def test_attach_volume_failure_when_volume_id_does_not_exist(self):
        """
        When L{attach_volume} is provided a C{volume_id} that doesn't
        exist, it logs an error and exits.
        """
        unit_name = "postgresql/0"
        instance_id = "i-123123"
        volume_id = "123-123-123"
        self._set_environment_vars({"JUJU_REMOTE_UNIT": unit_name})

        self.storage.load_environment = lambda: None
        self.storage.describe_volumes = lambda volume_id: {}

        result = self.assertRaises(
            SystemExit, self.storage.attach_volume, instance_id=instance_id,
            volume_id=volume_id)
        self.assertEqual(result.code, 1)
        message = ("Requested volume-id (%s) does not exist. Unable to "
                   "associate storage with %s" % (volume_id, unit_name))
        self.assertIn(
            message, util.hookenv._log_ERROR, "Not logged- %s" % message)

    def test_attach_volume_without_volume_label(self):
        """
        L{attach_volume} without a provided C{volume_label} or C{volume_id}
        will discover the nova volume id by searching L{describe_volumes}
        for volumes with a label based on the os.environ[JUJU_REMOTE_UNIT].
        """
        unit_name = "postgresql/0"
        volume_id = "123-123-123"
        instance_id = "i-123123123"
        volume_label = "%s unit volume" % unit_name
        self._set_environment_vars({"JUJU_REMOTE_UNIT": unit_name})
        self.storage.load_environment = lambda: None

        def mock_get_volume_id(label):
            self.assertEqual(label, volume_label)
            return volume_id
        self.storage.get_volume_id = mock_get_volume_id

        def mock_describe_volumes(my_id):
            self.assertEqual(my_id, volume_id)
            return {"status": "in-use", "device": "/dev/vdc"}
        self.storage.describe_volumes = mock_describe_volumes

        self.assertEqual(self.storage.attach_volume(instance_id), "/dev/vdc")
        message = (
            "Attaching %s (%s)" % (volume_label, volume_id))
        self.assertIn(
            message, util.hookenv._log_INFO, "Not logged- %s" % message)

    def test_attach_volume_when_volume_id_already_attached(self):
        """
        When L{attach_volume} is provided a C{volume_id} that already
        has the state C{in-use} it logs that the volume is already attached
        and returns.
        """
        unit_name = "postgresql/0"
        instance_id = "i-123123"
        volume_id = "123-123-123"
        self._set_environment_vars({"JUJU_REMOTE_UNIT": unit_name})

        self.storage.load_environment = lambda: None

        def mock_describe(my_id):
            self.assertEqual(my_id, volume_id)
            return {"status": "in-use", "device": "/dev/vdc"}
        self.storage.describe_volumes = mock_describe

        self.assertEqual(
            self.storage.attach_volume(instance_id, volume_id), "/dev/vdc")

        message = "Volume %s already attached. Done" % volume_id
        self.assertIn(
            message, util.hookenv._log_INFO, "Not logged- %s" % message)

    def test_attach_volume_when_volume_id_attaching_retry(self):
        """
        When L{attach_volume} is provided a C{volume_id} that has the status
        C{in-use} it logs, sleeps and retries until the volume is C{in-use}.
        """
        unit_name = "postgresql/0"
        instance_id = "i-123123"
        volume_id = "123-123-123"
        self._set_environment_vars({"JUJU_REMOTE_UNIT": unit_name})

        self.describe_count = 0

        self.storage.load_environment = lambda: None

        sleep = self.mocker.replace("time.sleep")
        sleep(5)
        self.mocker.replay()

        def mock_describe(my_id):
            self.assertEqual(my_id, volume_id)
            if self.describe_count == 0:
                self.describe_count += 1
                return {"status": "attaching"}
            else:
                self.describe_count += 1
                return {"status": "in-use", "device": "/dev/vdc"}
        self.storage.describe_volumes = mock_describe

        self.assertEqual(
            self.storage.attach_volume(instance_id, volume_id), "/dev/vdc")

        messages = ["Volume %s already attached. Done" % volume_id,
                    "Volume %s still attaching. Waiting." % volume_id]
        for message in messages:
            self.assertIn(
                message, util.hookenv._log_INFO, "Not logged- %s" % message)

    def test_attach_volume_failure_with_volume_unsupported_status(self):
        """
        When L{attach_volume} is provided a C{volume_id} that has an
        unsupported status. It logs the error and exits.
        """
        unit_name = "postgresql/0"
        instance_id = "i-123123"
        volume_id = "123-123-123"
        self._set_environment_vars({"JUJU_REMOTE_UNIT": unit_name})

        self.storage.load_environment = lambda: None

        def mock_describe(my_id):
            self.assertEqual(my_id, volume_id)
            return {"status": "deleting", "device": "/dev/vdc"}
        self.storage.describe_volumes = mock_describe

        result = self.assertRaises(
            SystemExit, self.storage.attach_volume, instance_id, volume_id)
        self.assertEqual(result.code, 1)
        message = ("Cannot attach volume. "
                   "Volume has unsupported status: deleting")
        self.assertIn(
            message, util.hookenv._log_ERROR, "Not logged- %s" % message)

    def test_attach_volume_creates_with_config_size(self):
        """
        When C{volume_id} is C{None}, L{attach_volume} will create a new
        volume with the configured C{default_volume_size} when the volume
        doesn't exist and C{size} is not provided.
        """
        unit_name = "postgresql/0"
        instance_id = "i-123123"
        volume_id = "123-123-123"
        volume_label = "%s unit volume" % unit_name
        default_volume_size = util.hookenv.config("default_volume_size")
        self._set_environment_vars({"JUJU_REMOTE_UNIT": unit_name})

        self.storage.load_environment = lambda: None
        self.storage.get_volume_id = lambda _: None

        def mock_describe(my_id):
            self.assertEqual(my_id, volume_id)
            return {"status": "in-use", "device": "/dev/vdc"}
        self.storage.describe_volumes = mock_describe

        def mock_nova_create(size, label, instance):
            self.assertEqual(size, default_volume_size)
            self.assertEqual(label, volume_label)
            self.assertEqual(instance, instance_id)
            return volume_id
        self.storage._nova_create_volume = mock_nova_create

        self.assertEqual(self.storage.attach_volume(instance_id), "/dev/vdc")
        message = "Attaching %s (%s)" % (volume_label, volume_id)
        self.assertIn(
            message, util.hookenv._log_INFO, "Not logged- %s" % message)

    def test_wb_nova_describe_volumes_command_error(self):
        """
        L{_nova_describe_volumes} will exit in error when the
        C{nova volume-list} command fails.
        """
        command = "nova volume-list"
        nova_list = self.mocker.replace(subprocess.check_output)
        nova_list(command, shell=True)
        self.mocker.throw(subprocess.CalledProcessError(1, command))
        self.mocker.replay()

        result = self.assertRaises(
            SystemExit, self.storage._nova_describe_volumes)
        self.assertEqual(result.code, 1)

        message = (
            "ERROR: Command '%s' returned non-zero exit status 1" % command)
        self.assertIn(
            message, util.hookenv._log_ERROR, "Not logged- %s" % message)

    def test_wb_nova_describe_volumes_without_attached_instances(self):
        """
        L{_nova_describe_volumes} parses the command to {nova volume-list} to
        create a C{dict} of volume information. When no C{instance_id}s are
        present the volumes are not attached so L{nova_volume_show} is not
        called to provide attachment details.
        """
        command = "nova volume-list"
        output = (
            "+-------------------------------------+\n"
            "| ID | Status | Display Name | Size | Volume Type| Attached to|\n"
            "+--------------------------------------------------------+\n"
            "| 123-123-123 | available | None           | 10 | None | |\n"
            "| 456-456-456 | available | my volume name | 8  | None | |\n"
            "+---------------------+----------------------------------+\n")

        nova_list = self.mocker.replace(subprocess.check_output)
        nova_list(command, shell=True)
        self.mocker.result(output)
        self.mocker.replay()

        def mock_nova_show(my_id):
            raise Exception("_nova_volume_show should not be called")
        self.storage._nova_volume_show = mock_nova_show

        expected = {"123-123-123": {"id": "123-123-123", "status": "available",
                                    "volume_label": "", "size": "10",
                                    "instance_id": "",
                                    "tags": {"volume_name": ""}},
                    "456-456-456": {"id": "456-456-456", "status": "available",
                                    "volume_label": "my volume name",
                                    "size": "8", "instance_id": "",
                                    "tags": {"volume_name": "my volume name"}}}
        self.assertEqual(self.storage._nova_describe_volumes(), expected)

    def test_wb_nova_describe_volumes_matches_volume_id_supplied(self):
        """
        L{_nova_describe_volumes} parses the command to {nova volume-list} to
        create a C{dict} of volume information. When C{volume_id} is provided
        return a C{dict} for the matched volume.
        """
        command = "nova volume-list"
        volume_id = "123-123-123"
        output = (
            "+-------------------------------------+\n"
            "| ID | Status | Display Name | Size | Volume Type| Attached to|\n"
            "+--------------------------------------------------------+\n"
            "| %s          | available | None           | 10 | None | |\n"
            "| 456-456-456 | available | my volume name | 8  | None | |\n"
            "+---------------------+----------------------------------+\n" %
            volume_id)

        nova_list = self.mocker.replace(subprocess.check_output)
        nova_list(command, shell=True)
        self.mocker.result(output)
        self.mocker.replay()

        expected = {"id": "123-123-123", "status": "available",
                    "volume_label": "", "size": "10",
                    "instance_id": "", "tags": {"volume_name": ""}}
        self.assertEqual(
            self.storage._nova_describe_volumes(volume_id), expected)

    def test_wb_nova_describe_volumes_unmatched_volume_id_supplied(self):
        """
        L{_nova_describe_volumes} parses the command to {nova volume-list} to
        create a C{dict} of volume information. Return and empty C{dict} when
        to volume matches the provided C{volume_id}.
        """
        command = "nova volume-list"
        volume_id = "123-123-123"
        output = (
            "+-------------------------------------+\n"
            "| ID | Status | Display Name | Size | Volume Type| Attached to|\n"
            "+--------------------------------------------------------+\n"
            "| 789-789-789 | available | None           | 10 | None | |\n"
            "| 456-456-456 | available | my volume name | 8  | None | |\n"
            "+---------------------+----------------------------------+\n")

        nova_list = self.mocker.replace(subprocess.check_output)
        nova_list(command, shell=True)
        self.mocker.result(output)
        self.mocker.replay()

        self.assertEqual(
            self.storage._nova_describe_volumes(volume_id), {})

    def test_wb_nova_describe_volumes_with_attached_instances(self):
        """
        L{_nova_describe_volumes} parses the command to {nova volume-list} to
        create a C{dict} of volume information. When C{instance_id}s are
        present the volumes are attached and L{nova_volume_show} will be called
        to provide attachment details.
        """
        command = "nova volume-list"
        attached_volume_id = "456-456-456"
        output = (
            "+-------------------------------------+\n"
            "| ID | Status | Display Name | Size | Volume Type| Attached to|\n"
            "+------------------------------------------------------------+\n"
            "| 123-123-123 | available | None    | 10 | None |          | |\n"
            "| %s          | in-use    | my name | 8  | None | i-789789 | |\n"
            "+---------------------+--------------------------------------+\n"
            % attached_volume_id)

        nova_list = self.mocker.replace(subprocess.check_output)
        nova_list(command, shell=True)
        self.mocker.result(output)
        self.mocker.replay()

        def mock_nova_show(my_id):
            self.assertEqual(my_id, attached_volume_id)
            return {
                "volume_label": "my name", "tags": {"volume_name": "my name"},
                "instance_id": "i-789789", "device": "/dev/vdx",
                "id": attached_volume_id, "status": "in-use",
                "availability_zone": "nova", "size": "8",
                "snapshot_id": "blah"}
        self.storage._nova_volume_show = mock_nova_show

        expected = {"123-123-123": {"id": "123-123-123", "status": "available",
                                    "volume_label": "", "size": "10",
                                    "instance_id": "",
                                    "tags": {"volume_name": ""}},
                    "456-456-456": {"id": "456-456-456", "status": "in-use",
                                    "volume_label": "my name",
                                    "size": "8", "instance_id": "i-789789",
                                    "snapshot_id": "blah",
                                    "device": "/dev/vdx",
                                    "availability_zone": "nova",
                                    "tags": {"volume_name": "my name"}}}
        self.assertEqual(self.storage._nova_describe_volumes(), expected)

    def test_wb_nova_create_volume(self):
        """
        L{_nova_create_volume} uses the command {nova volume-create} and
        logs its action.
        """
        instance_id = "i-123123"
        volume_id = "123-123-123"
        volume_label = "postgresql/0 unit volume"
        size = 10
        command = (
            "nova volume-create --display-name '%s' %s" % (volume_label, size))

        create = self.mocker.replace(subprocess.check_output)
        create(command, shell=True)
        self.mocker.replay()

        self.storage.get_volume_id = lambda label: volume_id

        self.assertEqual(
            self.storage._nova_create_volume(size, volume_label, instance_id),
            volume_id)
        message = (
            "Creating a %sGig volume named (%s) for instance %s" %
            (size, volume_label, instance_id))
        self.assertIn(
            message, util.hookenv._log_INFO, "Not logged- %s" % message)

    def test_wb_nova_create_volume_error_volume_not_created(self):
        """
        L{_nova_create_volume} will log an error and exit when unable to find
        the volume it just created.
        """
        instance_id = "i-123123"
        volume_label = "postgresql/0 unit volume"
        size = 10
        command = (
            "nova volume-create --display-name '%s' %s" % (volume_label, size))

        create = self.mocker.replace(subprocess.check_output)
        create(command, shell=True)
        self.mocker.replay()
        self.storage.get_volume_id = lambda label: None

        result = self.assertRaises(
            SystemExit, self.storage._nova_create_volume, size, volume_label,
            instance_id)
        self.assertEqual(result.code, 1)
        message = (
            "ERROR: Couldn't find newly created nova volume '%s'." %
            volume_label)
        self.assertIn(
            message, util.hookenv._log_ERROR, "Not logged- %s" % message)

    def test_wb_nova_create_volume_error_command_failed_with_retries(self):
        """
        L{_nova_create_volume} will log warnings and retry 3 times when an
        error is raised by the command  C{nova create-volume}. Upon failure of
        the third retry, L{_nova_create_volume} will log an error and exit 1.
        """
        instance_id = "i-123123"
        volume_label = "postgresql/0 unit volume"
        size = 10
        command = (
            "nova volume-create --display-name '%s' %s" % (volume_label, size))

        create = self.mocker.replace(subprocess.check_output)
        create(command, shell=True)
        self.mocker.count(4)
        self.mocker.throw(subprocess.CalledProcessError(1, command))
        sleep = self.mocker.replace("time.sleep")
        sleep(30)
        self.mocker.count(3)
        self.mocker.replay()

        result = self.assertRaises(
            SystemExit, self.storage._nova_create_volume, size, volume_label,
            instance_id)
        self.assertEqual(result.code, 1)
        self.assertEqual(len(util.hookenv._log_WARNING), 3)
        message = (
            "WARNING: Command '%s' returned non-zero exit status 1. "
            "Retrying 3 more times" % command)
        self.assertIn(
            message, util.hookenv._log_WARNING, "Not logged- %s" % message)

        message = (
            "ERROR: Command '%s' returned non-zero exit status 1." % command)
        self.assertIn(
            message, util.hookenv._log_ERROR, "Not logged- %s" % message)

    def test_wb_nova_attach_volume(self):
        """
        L{_nova_attach_volume} uses the command to {nova volume-attach} and
        returns the attached volume path.
        """
        instance_id = "i-123123"
        volume_id = "123-123-123"
        command = (
            "nova volume-attach %s %s auto | egrep -o \"/dev/vd[b-z]\"" %
            (instance_id, volume_id))

        attach = self.mocker.replace(subprocess.check_output)
        attach(command, shell=True)
        self.mocker.result("/dev/vdz\n")
        self.mocker.replay()

        self.assertEqual(
            self.storage._nova_attach_volume(instance_id, volume_id),
            "/dev/vdz")

    def test_wb_nova_attach_volume_no_device_path(self):
        """
        L{_nova_attach_volume} uses the command to {nova volume-attach} and
        returns an empty C{str} if the attached volume path was not discovered.
        """
        instance_id = "i-123123"
        volume_id = "123-123-123"
        command = (
            "nova volume-attach %s %s auto | egrep -o \"/dev/vd[b-z]\"" %
            (instance_id, volume_id))

        attach = self.mocker.replace(subprocess.check_output)
        attach(command, shell=True)
        self.mocker.result("\n")
        self.mocker.replay()

        self.assertEqual(
            self.storage._nova_attach_volume(instance_id, volume_id),
            "")

    def test_wb_nova_attach_volume_command_error_retries_three_times(self):
        """
        L{_nova_attach_volume} will warn on command error and retry the command
        3 times with sleeps in between. When the command C{nova volume-attach}
        exits in error on the third retry, an error is logged and the method
        exits 1.
        """
        instance_id = "i-123123"
        volume_id = "123-123-123"
        command = (
            "nova volume-attach %s %s auto | egrep -o \"/dev/vd[b-z]\"" %
            (instance_id, volume_id))
        attach = self.mocker.replace(subprocess.check_output)
        attach(command, shell=True)
        self.mocker.count(4)
        self.mocker.throw(subprocess.CalledProcessError(1, command))
        sleep = self.mocker.replace("time.sleep")
        sleep(30)
        self.mocker.count(3)
        self.mocker.replay()

        result = self.assertRaises(
            SystemExit, self.storage._nova_attach_volume, instance_id,
            volume_id)
        self.assertEqual(result.code, 1)
        self.assertEqual(len(util.hookenv._log_WARNING), 3)
        message = (
            "WARNING: Command '%s' returned non-zero exit status 1. "
            "Retrying 3 more times" % command)
        self.assertIn(
            message, util.hookenv._log_WARNING, "Not logged- %s" % message)
        message = (
            "ERROR: Command '%s' returned non-zero exit status 1." % command)
        self.assertIn(
            message, util.hookenv._log_ERROR, "Not logged- %s" % message)

    def test_detach_volume_no_volume_found(self):
        """
        When L{get_volume_id} is unable to find an attached volume and returns
        C{None}, L{detach_volume} will log a message and perform no work.
        """
        volume_label = "postgresql/0 unit volume"
        self.storage.load_environment = lambda: None

        def mock_get_volume_id(label):
            self.assertEqual(label, "postgresql/0 unit volume")
            return None
        self.storage.get_volume_id = mock_get_volume_id

        self.storage.detach_volume(volume_label)
        message = "Cannot find volume name to detach, done"
        self.assertIn(
            message, util.hookenv._log_INFO, "Not logged- %s" % message)

    def test_detach_volume_volume_already_detached(self):
        """
        When L{get_volume_id} finds a volume that is already C{available} it
        logs that the volume is already detached and does no work.
        """
        volume_label = "mycharm/1 unit volume"
        volume_id = "123-123-123"
        self.storage.load_environment = lambda: None

        def mock_get_volume_id(label):
            self.assertEqual(label, volume_label)
            return volume_id
        self.storage.get_volume_id = mock_get_volume_id

        def mock_describe_volumes(my_id):
            self.assertEqual(my_id, volume_id)
            return {"status": "available"}
        self.storage.describe_volumes = mock_describe_volumes

        self.storage.detach_volume(volume_label)  # pass in our volume_label
        message = "Volume (%s) already detached. Done" % volume_id
        self.assertIn(
            message, util.hookenv._log_INFO, "Not logged- %s" % message)

    def test_detach_volume_command_error(self):
        """
        When the C{nova volume-detach} command fails, L{detach_volume} will
        log a message and exit in error.
        """
        instance_id = "i-123123"
        volume_id = "123-123-123"
        volume_label = "mycharm/1 unit volume"
        self.storage.load_environment = lambda: None

        def mock_get_volume_id(label):
            self.assertEqual(label, volume_label)
            return volume_id
        self.storage.get_volume_id = mock_get_volume_id

        def mock_describe_volumes(my_id):
            self.assertEqual(my_id, volume_id)
            return {"status": "in-use", "instance_id": instance_id}
        self.storage.describe_volumes = mock_describe_volumes

        command = "nova volume-detach %s %s" % (instance_id, volume_id)
        nova_cmd = self.mocker.replace(subprocess.check_call)
        nova_cmd(command, shell=True)
        self.mocker.throw(subprocess.CalledProcessError(1, command))
        self.mocker.replay()

        result = self.assertRaises(
            SystemExit, self.storage.detach_volume, volume_label)
        self.assertEqual(result.code, 1)
        message = (
            "ERROR: Couldn't detach volume. Command '%s' "
            "returned non-zero exit status 1" % command)
        self.assertIn(
            message, util.hookenv._log_ERROR, "Not logged- %s" % message)

    def test_detach_volume(self):
        """
        When L{get_volume_id} finds a volume associated with this instance
        which has a volume state not equal to C{available}, it detaches that
        volume using nova commands.
        """
        volume_id = "123-123-123"
        instance_id = "i-123123"
        volume_label = "postgresql/0 unit volume"
        self.storage.load_environment = lambda: None

        def mock_get_volume_id(label):
            self.assertEqual(label, volume_label)
            return volume_id
        self.storage.get_volume_id = mock_get_volume_id

        def mock_describe_volumes(my_id):
            self.assertEqual(my_id, volume_id)
            return {"status": "in-use", "instance_id": instance_id}
        self.storage.describe_volumes = mock_describe_volumes

        command = "nova volume-detach %s %s" % (instance_id, volume_id)
        nova_cmd = self.mocker.replace(subprocess.check_call)
        nova_cmd(command, shell=True)
        self.mocker.replay()

        self.storage.detach_volume(volume_label)
        message = (
            "Detaching volume (%s) from instance %s" %
            (volume_id, instance_id))
        self.assertIn(
            message, util.hookenv._log_INFO, "Not logged- %s" % message)


class MockBotoEC2Connection(object):
    """Mock all calls to python-boto's EC2Connection class."""
    def __init__(self, get_volumes=None, get_instances=None,
                 create_volume=None, create_tags=None, attach_volume=None,
                 detach_volume=None):
        self.get_all_volumes = get_volumes
        self.get_all_instances = get_instances
        self.create_volume = create_volume
        self.attach_volume = attach_volume
        self.detach_volume = detach_volume
        self.create_tags = create_tags


class MockEucaCommand(object):
    def __init__(self, result):
        self.result = result

    def main(self):
        return self.result


class MockEucaReservation(object):
    def __init__(self, instances):
        self.instances = instances
        self.id = 1


class MockEucaInstance(object):
    def __init__(self, instance_id=None, ip_address=None, image_id=None,
                 instance_type=None, kernel=None, private_dns_name=None,
                 public_dns_name=None, state=None, tags=[],
                 availability_zone=None):
        self.id = instance_id
        self.ip_address = ip_address
        self.image_id = image_id
        self.instance_type = instance_type
        self.kernel = kernel
        self.private_dns_name = private_dns_name
        self.public_dns_name = public_dns_name
        self.state = state
        self.tags = tags
        self.placement = availability_zone


class MockAttachData(object):
    def __init__(self, device, instance_id):
        self.device = device
        self.instance_id = instance_id


class MockVolume(object):
    def __init__(self, vol_id, device=None, instance_id=None, zone=None,
                 size=None, status=None, snapshot_id=None, tags=None):
        self.id = vol_id
        self.attach_data = MockAttachData(device, instance_id)
        self.zone = zone
        self.size = size
        self.status = status
        self.snapshot_id = snapshot_id
        self.tags = tags


class TestEC2Util(mocker.MockerTestCase):

    def setUp(self):
        super(TestEC2Util, self).setUp()
        self.maxDiff = None
        util.hookenv = TestHookenv(
            {"key": "ec2key", "secret": "ec2password",
             "endpoint": "https://ec2-region-1.com",
             "default_volume_size": 11})
        util.EC2_BOTO_CONFIG_FILE = self.makeFile()
        util.log = util.hookenv.log
        self.storage = StorageServiceUtil("ec2")

    def _set_environment_vars(self, environ={}):
        self.addCleanup(setattr, util.os, "environ", util.os.environ)
        util.os.environ = environ

    def test_wb_setup_boto_config(self):
        """
        L{_setup_boto_config} writes C{EC2_BOTO_CONFIG_FILE} with the provided
        C{key} and C{secret} parameters.
        """
        key = "my_ec2_key"
        secret = "my_secret_access_key"
        expected = ["[Credentials]",
                    "aws_access_key_id = %s" % key,
                    "aws_secret_access_key = %s" % secret]
        self.storage._setup_boto_config(key=key, secret=secret)
        with open(util.EC2_BOTO_CONFIG_FILE) as boto_config:
            lines = boto_config.read().splitlines()
        self.assertEqual(lines, expected)

    def test_load_environment_with_ec2_variables(self):
        """
        L{load_environment} will setup script environment variables for ec2
        by mapping configuration values provided to openstack OS_* environment
        variables and then call L{validate_credentials} to assert
        that environment variables provided give access to the service.
        """
        self._set_environment_vars({})

        def mock_validate():
            pass
        self.storage.validate_credentials = mock_validate

        self.storage.load_environment()
        expected = {
            "EC2_ACCESS_KEY": "ec2key",
            "EC2_SECRET_KEY": "ec2password",
            "EC2_URL": "https://ec2-region-1.com"
        }
        self.assertEqual(util.os.environ, expected)

    def test_load_environment_error_missing_config_options(self):
        """
        L{load_environment} will exit in failure and log a message if any
        required configuration option is not set.
        """
        def mock_validate():
            raise SystemExit("something invalid")
        self.storage.validate_credentials = mock_validate

        self.assertRaises(SystemExit, self.storage.load_environment)

    def test_validate_credentials(self):
        """
        L{validate_credentials} will succeed when a boto's L{get_all_volumes}
        succeeds using C{EC2_URL} environment variable from charm configuration
        options.
        """
        ec2_url = "https://ec2.us-west-1.amazonaws.com"
        self._set_environment_vars({"EC2_URL": ec2_url})
        connect = self.mocker.replace("boto.ec2.connect_to_region")
        connect("us-west-1")

        def get_volumes():
            return []

        self.mocker.result(MockBotoEC2Connection(
            get_volumes=get_volumes))
        self.mocker.replay()

        self.storage.validate_credentials()
        message = (
            "Validated charm configuration credentials have access to "
            "block storage service"
        )
        self.assertIn(
            message, util.hookenv._log_INFO, "Not logged- %s" % message)

    def test_get_volume_id_by_volume_name(self):
        """
        L{get_volume_id} provided with a existing C{volume_name} returns the
        corresponding ec2 volume id from L{_ec2_describe_volumes}.
        """
        volume_name = "my-volume"
        volume_id = "12312412-412312"

        def mock_describe(val):
            self.assertIsNone(val)
            return {volume_id: {"tags": {"volume_name": volume_name}},
                    "456456-456456": {"tags": {"volume_name": "blah"}}}
        self.storage._ec2_describe_volumes = mock_describe

        self.assertEqual(self.storage.get_volume_id(volume_name), volume_id)

    def test_get_volume_id_without_volume_name(self):
        """
        L{get_volume_id} without a provided C{volume_name} will discover the
        nova volume id by searching L{_ec2_describe_volumes} for volumes
        labelled with the os.environ[JUJU_REMOTE_UNIT].
        """
        unit_name = "postgresql/0"
        self._set_environment_vars({"JUJU_REMOTE_UNIT": unit_name})
        volume_id = "123134124-1241412-1242141"

        def mock_describe(val):
            self.assertIsNone(val)
            return {volume_id:
                    {"tags": {"volume_name": "postgresql/0 unit volume"}},
                    "456456-456456": {"tags": {"volume_name": "blah"}}}
        self.storage._ec2_describe_volumes = mock_describe

        self.assertEqual(self.storage.get_volume_id(), volume_id)

    def test_get_volume_id_without_volume_name_no_matching_volume(self):
        """
        L{get_volume_id} without a provided C{volume_name} will return C{None}
        when it cannot find a matching volume label from
        L{_ec2_describe_volumes} for the os.environ[JUJU_REMOTE_UNIT].
        """
        unit_name = "postgresql/0"
        self._set_environment_vars({"JUJU_REMOTE_UNIT": unit_name})

        def mock_describe(val):
            self.assertIsNone(val)
            return {"123123-123123":
                    {"tags": {"volume_name": "postgresql/1 unit volume"}},
                    "456456-456456": {"tags": {"volume_name": "blah"}}}
        self.storage._ec2_describe_volumes = mock_describe

        self.assertIsNone(self.storage.get_volume_id())

    def test_get_volume_id_without_volume_name_multiple_matching_volumes(self):
        """
        L{get_volume_id} does not support multiple volumes associated with the
        the instance represented by os.environ[JUJU_REMOTE_UNIT]. When
        C{volume_name} is not specified and L{_ec2_describe_volumes} returns
        multiple results the function exits with an error.
        """
        unit_name = "postgresql/0"
        self._set_environment_vars({"JUJU_REMOTE_UNIT": unit_name})

        def mock_describe(val):
            self.assertIsNone(val)
            return {"123123-123123":
                    {"tags": {"volume_name": "postgresql/0 unit volume"}},
                    "456456-456456":
                    {"tags": {"volume_name": "unit postgresql/0 volume2"}}}
        self.storage._ec2_describe_volumes = mock_describe

        result = self.assertRaises(SystemExit, self.storage.get_volume_id)
        self.assertEqual(result.code, 1)
        message = (
            "Error: Multiple volumes are associated with %s. "
            "Cannot get_volume_id." % unit_name)
        self.assertIn(
            message, util.hookenv._log_ERROR, "Not logged- %s" % message)

    def test_attach_volume_failure_when_volume_id_does_not_exist(self):
        """
        When L{attach_volume} is provided a C{volume_id} that doesn't
        exist, it logs an error and exits.
        """
        unit_name = "postgresql/0"
        instance_id = "i-123123"
        volume_id = "123-123-123"
        self._set_environment_vars({"JUJU_REMOTE_UNIT": unit_name})

        self.storage.load_environment = lambda: None
        self.storage._ec2_describe_volumes = lambda volume_id: {}

        result = self.assertRaises(
            SystemExit, self.storage.attach_volume, instance_id=instance_id,
            volume_id=volume_id)
        self.assertEqual(result.code, 1)
        message = ("Requested volume-id (%s) does not exist. Unable to "
                   "associate storage with %s" % (volume_id, unit_name))
        self.assertIn(
            message, util.hookenv._log_ERROR, "Not logged- %s" % message)

    def test_attach_volume_without_volume_label(self):
        """
        L{attach_volume} without a provided C{volume_label} or C{volume_id}
        will discover the nova volume id by searching L{_ec2_describe_volumes}
        for volumes with a label based on the os.environ[JUJU_REMOTE_UNIT].
        """
        unit_name = "postgresql/0"
        volume_id = "123-123-123"
        instance_id = "i-123123123"
        volume_label = "%s unit volume" % unit_name
        self._set_environment_vars({"JUJU_REMOTE_UNIT": unit_name})
        self.storage.load_environment = lambda: None

        def mock_get_volume_id(label):
            self.assertEqual(label, volume_label)
            return volume_id
        self.storage.get_volume_id = mock_get_volume_id

        def mock_describe_volumes(my_id):
            self.assertEqual(my_id, volume_id)
            return {"status": "in-use", "device": "/dev/vdc"}
        self.storage._ec2_describe_volumes = mock_describe_volumes

        self.assertEqual(self.storage.attach_volume(instance_id), "/dev/vdc")
        message = (
            "Attaching %s (%s)" % (volume_label, volume_id))
        self.assertIn(
            message, util.hookenv._log_INFO, "Not logged- %s" % message)

    def test_attach_volume_when_volume_id_already_attached(self):
        """
        When L{attach_volume} is provided a C{volume_id} that already
        has the state C{in-use} it logs that the volume is already attached
        and returns.
        """
        unit_name = "postgresql/0"
        instance_id = "i-123123"
        volume_id = "123-123-123"
        self._set_environment_vars({"JUJU_REMOTE_UNIT": unit_name})

        self.storage.load_environment = lambda: None

        def mock_describe(my_id):
            self.assertEqual(my_id, volume_id)
            return {"status": "in-use", "device": "/dev/vdc"}
        self.storage._ec2_describe_volumes = mock_describe

        self.assertEqual(
            self.storage.attach_volume(instance_id, volume_id), "/dev/vdc")

        message = "Volume %s already attached. Done" % volume_id
        self.assertIn(
            message, util.hookenv._log_INFO, "Not logged- %s" % message)

    def test_attach_volume_failure_with_volume_unsupported_status(self):
        """
        When L{attach_volume} is provided a C{volume_id} that has an
        unsupported status. It logs the error and exits.
        """
        unit_name = "postgresql/0"
        instance_id = "i-123123"
        volume_id = "123-123-123"
        self._set_environment_vars({"JUJU_REMOTE_UNIT": unit_name})

        self.storage.load_environment = lambda: None

        def mock_describe(my_id):
            self.assertEqual(my_id, volume_id)
            return {"status": "deleting", "device": "/dev/vdc"}
        self.storage._ec2_describe_volumes = mock_describe

        result = self.assertRaises(
            SystemExit, self.storage.attach_volume, instance_id, volume_id)
        self.assertEqual(result.code, 1)
        message = ("Cannot attach volume. "
                   "Volume has unsupported status: deleting")
        self.assertIn(
            message, util.hookenv._log_ERROR, "Not logged- %s" % message)

    def test_attach_volume_creates_with_config_size(self):
        """
        When C{volume_id} is C{None}, L{attach_volume} will create a new
        volume with the configured C{default_volume_size} when the volume
        doesn't exist and C{size} is not provided.
        """
        unit_name = "postgresql/0"
        instance_id = "i-123123"
        volume_id = "123-123-123"
        volume_label = "%s unit volume" % unit_name
        default_volume_size = util.hookenv.config("default_volume_size")
        self._set_environment_vars({"JUJU_REMOTE_UNIT": unit_name})

        self.storage.load_environment = lambda: None
        self.storage.get_volume_id = lambda _: None

        def mock_describe(my_id):
            self.assertEqual(my_id, volume_id)
            return {"status": "in-use", "device": "/dev/vdc"}
        self.storage._ec2_describe_volumes = mock_describe

        def mock_ec2_create(size, label, instance):
            self.assertEqual(size, default_volume_size)
            self.assertEqual(label, volume_label)
            self.assertEqual(instance, instance_id)
            return volume_id
        self.storage._ec2_create_volume = mock_ec2_create

        self.assertEqual(self.storage.attach_volume(instance_id), "/dev/vdc")
        message = "Attaching %s (%s)" % (volume_label, volume_id)
        self.assertIn(
            message, util.hookenv._log_INFO, "Not logged- %s" % message)

    def test_wb_ec2_validate_credentials_invalid_ec2_url_environment_var(self):
        """
        L{_ec2_validate_credentials} will exit in error when the C{EC2_URL}
        environment variable is invalid.
        """
        self._set_environment_vars({"EC2_URL": "not-a-valid-ec2-url"})

        result = self.assertRaises(
            SystemExit, self.storage._ec2_validate_credentials)
        self.assertEqual(result.code, 1)
        message = (
            "ERROR: Couldn't get region from EC2_URL environment variable: "
            "not-a-valid-ec2-url.")
        self.assertIn(
            message, util.hookenv._log_ERROR, "Not logged- %s" % message)

    def test_wb_ec2_validate_credentials_absent_ec2_url_environment_var(self):
        """
        L{_ec2_validate_credentials} will exit in error when the C{EC2_URL}
        environment variable is not present.
        """
        self._set_environment_vars({})

        result = self.assertRaises(
            SystemExit, self.storage._ec2_validate_credentials)
        self.assertEqual(result.code, 1)
        message = (
            "ERROR: Couldn't get region from EC2_URL environment variable: "
            "NOT SET.")
        self.assertIn(
            message, util.hookenv._log_ERROR, "Not logged- %s" % message)

    def test_wb_ec2_validate_credentials_unavailable_credentials(self):
        """
        L{_ec2_validate_credentials} will exit in error when the EC2
        credentials are not present in /etc/boto.cfg.
        """
        self._set_environment_vars(
            {"EC2_URL": "https://ec2.us-west-1.amazonaws.com"})
        connect = self.mocker.replace("boto.ec2.connect_to_region")
        connect("us-west-1")

        def get_volumes_error():
            raise NoAuthHandlerFound("No handler was ready to auth.")

        self.mocker.result(
            MockBotoEC2Connection(get_volumes=get_volumes_error))
        self.mocker.replay()

        result = self.assertRaises(
            SystemExit, self.storage._ec2_validate_credentials)
        self.assertEqual(result.code, 1)
        message = (
            "ERROR: EC2 credentials not found in /etc/boto.cfg. Cannot "
            "authenticate.")
        self.assertIn(
            message, util.hookenv._log_ERROR, "Not logged- %s" % message)

    def test_wb_ec2_validate_credentials_invalid_credentials(self):
        """
        L{_ec2_validate_credentials} will exit in error when the EC2
        credentials are not valid for the C{EC2_URL} provided.
        """
        self._set_environment_vars(
            {"EC2_URL": "https://ec2.us-west-1.amazonaws.com"})
        connect = self.mocker.replace("boto.ec2.connect_to_region")
        connect("us-west-1")

        def get_volumes_error():
            raise EC2ResponseError(401, "Unauthorized")

        self.mocker.result(
            MockBotoEC2Connection(get_volumes=get_volumes_error))
        self.mocker.replay()

        result = self.assertRaises(
            SystemExit, self.storage._ec2_validate_credentials)
        self.assertEqual(result.code, 1)
        message = (
            "ERROR: Invalid EC2 credentials in /etc/boto.cfg. Unauthorized.")
        self.assertIn(
            message, util.hookenv._log_ERROR, "Not logged- %s" % message)

    def test_wb_ec2_validate_credentials_timeout(self):
        """
        L{_ec2_validate_credentials} will exit in error when a connection
        timeout occurs against the region endpoint.
        """
        ec2_url = "https://ec2.us-west-1.amazonaws.com"
        self._set_environment_vars({"EC2_URL": ec2_url})
        connect = self.mocker.replace("boto.ec2.connect_to_region")
        connect("us-west-1")

        def get_volumes_error():
            raise gaierror("Temporary failure in name resolution")

        self.mocker.result(
            MockBotoEC2Connection(get_volumes=get_volumes_error))
        self.mocker.replay()

        result = self.assertRaises(
            SystemExit, self.storage._ec2_validate_credentials)
        self.assertEqual(result.code, 1)
        message = "ERROR: Connectivity error accessing %s" % ec2_url
        self.assertIn(
            message, util.hookenv._log_ERROR, "Not logged- %s" % message)

    def test_wb_ec2_describe_volumes_without_attached_instances(self):
        """
        L{_ec2_describe_volumes} parses the results of boto.ec2.get_all_volumes
        C{DescribeVolumes} to create a C{dict} of volume information. When no
        C{instance_id}s are present the volumes are not attached so no
        C{device} or C{instance_id} information will be present.
        """
        volume1 = MockVolume(
            "123-123-123", device="/dev/notshown", instance_id="notseen",
            zone="ec2-az1", size="10", status="available",
            snapshot_id="some-shot", tags={})
        volume2 = MockVolume(
            "456-456-456", device="/dev/notshown", instance_id="notseen",
            zone="ec2-az2", size="8", status="available",
            snapshot_id="some-shot", tags={"volume_name": "my volume name"})

        def get_volumes():
            return [volume1, volume2]

        self.storage.ec2_conn = MockBotoEC2Connection(get_volumes=get_volumes)

        expected = {"123-123-123": {"id": "123-123-123", "status": "available",
                                    "device": "",
                                    "availability_zone": "ec2-az1",
                                    "volume_label": "", "size": "10",
                                    "instance_id": "",
                                    "snapshot_id": "some-shot",
                                    "tags": {"volume_name": ""}},
                    "456-456-456": {"id": "456-456-456", "status": "available",
                                    "device": "",
                                    "availability_zone": "ec2-az2",
                                    "volume_label": "my volume name",
                                    "size": "8", "instance_id": "",
                                    "snapshot_id": "some-shot",
                                    "tags": {"volume_name": "my volume name"}}}
        self.assertEqual(self.storage._ec2_describe_volumes(), expected)

    def test_wb_ec2_describe_volumes_matches_volume_id_supplied(self):
        """
        L{_ec2_describe_volumes} matches the provided C{volume_id} to the
        results of boto's L{get_all_volumes}.
        """
        volume_id = "123-123-123"
        volume1 = MockVolume(
            volume_id, device="/dev/notshown", instance_id="notseen",
            zone="ec2-az1", size="10", status="available",
            snapshot_id="some-shot", tags={})
        volume2 = MockVolume(
            "456-456-456", device="/dev/notshown", instance_id="notseen",
            zone="ec2-az2", size="8", status="available",
            snapshot_id="some-shot", tags={"volume_name": "my volume name"})

        def get_volumes():
            return [volume1, volume2]

        self.storage.ec2_conn = MockBotoEC2Connection(get_volumes=get_volumes)

        expected = {
            "id": volume_id, "status": "available", "device": "",
            "availability_zone": "ec2-az1", "volume_label": "", "size": "10",
            "instance_id": "", "snapshot_id": "some-shot",
            "tags": {"volume_name": ""}}
        self.assertEqual(
            self.storage._ec2_describe_volumes(volume_id), expected)

    def test_wb_ec2_describe_volumes_unmatched_volume_id_supplied(self):
        """
        L{_ec2_describe_volumes} returns an empty C{dict} when it does not
        match the provided C{volume_id} to any volumes reported by boto's
        L{get_all_volumes} method.
        """
        unmatched_volume_id = "456-456-456"
        volume1 = MockVolume(
            "123-123-123", device="/dev/notshown", instance_id="notseen",
            zone="ec2-az1", size="10", status="available",
            snapshot_id="some-shot", tags={})

        def get_volumes():
            return [volume1]

        self.storage.ec2_conn = MockBotoEC2Connection(get_volumes=get_volumes)

        self.assertEqual(
            self.storage._ec2_describe_volumes(unmatched_volume_id), {})

    def test_wb_ec2_describe_volumes_with_attached_instances(self):
        """
        L{_ec2_describe_volumes} will report attached instance_id information
        when boto's L{get_all_volumes} lists instance information for a volume.
        """
        volume1 = MockVolume(
            "123-123-123", device="/dev/notshown", instance_id="notseen",
            zone="ec2-az1", size="10", status="available",
            snapshot_id="some-shot", tags={})
        volume2 = MockVolume(
            "456-456-456", device="/dev/xvdc", instance_id="i-456456",
            zone="ec2-az2", size="8", status="in-use",
            snapshot_id="some-shot", tags={"volume_name": "my volume name"})

        def get_volumes():
            return [volume1, volume2]

        self.storage.ec2_conn = MockBotoEC2Connection(get_volumes=get_volumes)

        expected = {"123-123-123": {"id": "123-123-123", "status": "available",
                                    "device": "",
                                    "availability_zone": "ec2-az1",
                                    "volume_label": "", "size": "10",
                                    "instance_id": "",
                                    "snapshot_id": "some-shot",
                                    "tags": {"volume_name": ""}},
                    "456-456-456": {"id": "456-456-456", "status": "in-use",
                                    "device": "/dev/xvdc",
                                    "availability_zone": "ec2-az2",
                                    "volume_label": "my volume name",
                                    "size": "8", "instance_id": "i-456456",
                                    "snapshot_id": "some-shot",
                                    "tags": {"volume_name": "my volume name"}}}
        self.assertEqual(
            self.storage._ec2_describe_volumes(), expected)

    def test_wb_ec2_create_volume(self):
        """
        L{_ec2_create_volume} calls boto's L{create_volume} to create
        a volume. It determines the availability zone for the volume by
        querying L{_ec2_describe_instances} for the provided C{instance_id}
        to ensure the volume is created in the same availablity zone.
        L{_ec2_create_volume} will also call L{_ec2_create_tag} to setup the
        C{volume_name} tag for the volume.
        """
        instance_id = "i-123123"
        volume_id = "123-123-123"
        volume_label = "postgresql/0 unit volume"
        size = 10
        zone = "ec2-az3"

        reservation = MockEucaReservation(
            [MockEucaInstance(
                instance_id=instance_id, availability_zone=zone)])

        def get_instances():
            return [reservation]

        def create_volume(size, zone):
            self.assertEqual(zone, "ec2-az3")
            self.assertEqual(size, 10)
            return MockVolume(vol_id=volume_id)

        def create_tags(resource_ids, tags):
            self.assertEqual(resource_ids, [volume_id])
            self.assertEqual({"volume_name": volume_label}, tags)

        self.storage.ec2_conn = MockBotoEC2Connection(
            get_instances=get_instances, create_volume=create_volume,
            create_tags=create_tags)

        self.assertEqual(
            self.storage._ec2_create_volume(size, volume_label, instance_id),
            volume_id)
        message = (
            "Creating a %sGig volume named (%s) for instance %s" %
            (size, volume_label, instance_id))
        self.assertIn(
            message, util.hookenv._log_INFO, "Not logged- %s" % message)

    def test_wb_ec2_create_volume_error_different_region_configured(self):
        """
        When L{_ec2_create_volume} received an C{instance_id} which is not
        present in the current output of C{ec2_describe_instances} it is likely
        that the region for which the charm was configured is not the region in
        which the block-storage-broker is deployed. Log an error suggesting
        the misconfiguration of the charm in this case.
        """
        instance_id = "i-123123"
        volume_label = "postgresql/0 unit volume"
        size = 10

        def get_instances():
            return []

        self.storage.ec2_conn = MockBotoEC2Connection(
            get_instances=get_instances)

        result = self.assertRaises(
            SystemExit, self.storage._ec2_create_volume, size, volume_label,
            instance_id)
        self.assertEqual(result.code, 1)
        config = util.hookenv.config_get()
        message = (
            "ERROR: Could not create volume for instance %s. No instance "
            "details discovered by boto.ec2.get_all_instances. Maybe the "
            "charm configured endpoint %s is not valid for this region." %
            (instance_id, config["endpoint"]))
        self.assertIn(
            message, util.hookenv._log_ERROR, "Not logged- %s" % message)

    def test_wb_ec2_create_volume_error(self):
        """
        L{_ec2_create_volume} will log an error and exit when it receives an
        exception from boto's C{create_volume} command.
        """
        instance_id = "i-123123"
        volume_label = "postgresql/0 unit volume"
        size = 10
        zone = "ec2-az3"

        reservation = MockEucaReservation(
            [MockEucaInstance(
                instance_id=instance_id, availability_zone=zone)])

        def get_instances():
            return [reservation]

        def create_volume(size, zone):
            self.assertEqual(zone, "ec2-az3")
            self.assertEqual(size, 10)
            raise EC2ResponseError(401, "Unauthorized")

        self.storage.ec2_conn = MockBotoEC2Connection(
            get_instances=get_instances, create_volume=create_volume)

        result = self.assertRaises(
            SystemExit, self.storage._ec2_create_volume, size, volume_label,
            instance_id)
        self.assertEqual(result.code, 1)
        message = "ERROR: EC2ResponseError: 401 Unauthorized\n"
        self.assertIn(
            message, util.hookenv._log_ERROR, "Not logged- %s" % message)

    def test_wb_ec2_attach_volume(self):
        """
        L{_ec2_attach_volume} uses the EC2 boto's C{attach-volume} and
        returns the attached device path.
        """
        instance_id = "i-123123"
        volume_id = "123-123-123"
        device = "/dev/xvdc"

        def attach_volume(volume_id, instance_id, device):
            self.assertEqual(instance_id, "i-123123")
            self.assertEqual(volume_id, "123-123-123")
            self.assertEqual(device, "/dev/xvdc")

        self.storage.ec2_conn = MockBotoEC2Connection(
            attach_volume=attach_volume)

        self.assertEqual(
            self.storage._ec2_attach_volume(instance_id, volume_id),
            device)

    def test_wb_ec2_attach_volume_failed(self):
        """
        L{_ec2_attach_volume} exits in error when boto's C{attach-volume}
        fails.
        """
        instance_id = "i-123123"
        volume_id = "123-123-123"

        def attach_volume(volume_id, instance_id, device):
            raise EC2ResponseError(401, "Unauthorized")

        self.storage.ec2_conn = MockBotoEC2Connection(
            attach_volume=attach_volume)

        result = self.assertRaises(
            SystemExit, self.storage._ec2_attach_volume, instance_id,
            volume_id)
        self.assertEqual(result.code, 1)
        message = "ERROR: EC2ResponseError: 401 Unauthorized\n"
        self.assertIn(
            message, util.hookenv._log_ERROR, "Not logged- %s" % message)

    def test_detach_volume_no_volume_found(self):
        """
        When L{get_volume_id} is unable to find an attached volume and returns
        C{None}, L{detach_volume} will log a message and perform no work.
        """
        volume_label = "postgresql/0 unit volume"
        self.storage.load_environment = lambda: None

        def mock_get_volume_id(label):
            self.assertEqual(label, volume_label)
            return None
        self.storage.get_volume_id = mock_get_volume_id

        self.storage.detach_volume(volume_label)
        message = "Cannot find volume name to detach, done"
        self.assertIn(
            message, util.hookenv._log_INFO, "Not logged- %s" % message)

    def test_detach_volume_volume_already_detached(self):
        """
        When L{get_volume_id} finds a volume that is already C{available} it
        logs that the volume is already detached and does no work.
        """
        volume_label = "postgresql/0 unit volume"
        volume_id = "123-123-123"
        self.storage.load_environment = lambda: None

        def mock_get_volume_id(label):
            self.assertEqual(label, volume_label)
            return volume_id
        self.storage.get_volume_id = mock_get_volume_id

        def mock_describe_volumes(my_id):
            self.assertEqual(my_id, volume_id)
            return {"status": "available"}
        self.storage.describe_volumes = mock_describe_volumes

        self.storage.detach_volume(volume_label)
        message = "Volume (%s) already detached. Done" % volume_id
        self.assertIn(
            message, util.hookenv._log_INFO, "Not logged- %s" % message)

    def test_detach_volume(self):
        """
        When L{get_volume_id} finds a volume associated with this instance
        which has a volume state not equal to C{available}, it detaches that
        volume using boto's L{detach_volume}.
        """
        volume_label = "postgresql/0 unit volume"
        volume_id = "123-123-123"
        instance_id = "i-123123"
        self.storage.load_environment = lambda: None

        def mock_get_volume_id(label):
            self.assertEqual(label, volume_label)
            return volume_id
        self.storage.get_volume_id = mock_get_volume_id

        def mock_describe_volumes(my_id):
            self.assertEqual(my_id, volume_id)
            return {"status": "in-use", "instance_id": instance_id}
        self.storage.describe_volumes = mock_describe_volumes

        def detach_volume(instance_id, volume_id):
            self.assertEqual(instance_id, "i-123123")
            self.assertEqual(volume_id, "123-123-123")

        self.storage.ec2_conn = MockBotoEC2Connection(
            detach_volume=detach_volume)

        self.storage.detach_volume(volume_label)
        message = (
            "Detaching volume (%s) from instance %s" %
            (volume_id, instance_id))
        self.assertIn(
            message, util.hookenv._log_INFO, "Not logged- %s" % message)

    def test_detach_volume_error(self):
        """
        An error is logged and we exit(1) when boto's L{detach_volume} raises
        an error.
        """
        volume_label = "postgresql/0 unit volume"
        volume_id = "123-123-123"
        instance_id = "i-123123"
        self.storage.load_environment = lambda: None

        def mock_get_volume_id(label):
            self.assertEqual(label, volume_label)
            return volume_id
        self.storage.get_volume_id = mock_get_volume_id

        def mock_describe_volumes(my_id):
            self.assertEqual(my_id, volume_id)
            return {"status": "in-use", "instance_id": instance_id}
        self.storage.describe_volumes = mock_describe_volumes

        def detach_volume_error(instance_id, volume_id):
            raise EC2ResponseError(401, "Unauthorized")

        self.storage.ec2_conn = MockBotoEC2Connection(
            detach_volume=detach_volume_error)

        result = self.assertRaises(
            SystemExit, self.storage.detach_volume, volume_label)
        self.assertEqual(result.code, 1)
        message = (
            "ERROR: Couldn't detach volume. EC2ResponseError: "
            "401 Unauthorized\n")
        self.assertIn(
            message, util.hookenv._log_ERROR, "Not logged- %s" % message)
